"""
groupme-bot

A simple bot builder for GroupMe
"""
from .application import Application
from .attachment import (
    ImageAttachment, LocationAttachment, SplitAttachment, EmojiAttachment, MentionsAttachment, parse_attachment
)
from .bot import Bot, Context
from .callback import Callback
from .groupme import GroupMe

__version__ = "0.2.12"
__author__ = "Branden Colen"
__all__ = [
    "Application",
    "Bot",
    "Callback",
    "Context",
    "EmojiAttachment",
    "ImageAttachment",
    "LocationAttachment",
    "MentionsAttachment",
    "SplitAttachment",
    "parse_attachment",
    "GroupMe"
]



from typing import List

from .attachment import Attachment, parse_attachment


class Callback(object):
    __slots__ = '_callback_dict'

    def __init__(self, callback_dict: dict):
        """
        The callback provided by GroupMe. When any message is sent to the group, GroupMe sends a callback to the
        bots endpoint. This object is generated from the callback JSON to allow for easy use by the bot handlers.
        :param dict callback_dict: A dict of the JSON body extracted from the GroupMe callback POST request
        """
        self._callback_dict: dict = callback_dict

    @property
    def attachments(self) -> List[Attachment]:
        attachments = self._callback_dict.get("attachments")
        if attachments is None:
            return []
        return [parse_attachment(attachment) for attachment in attachments]

    @property
    def avatar_url(self) -> str:
        return self._callback_dict.get("avatar_url")

    @property
    def created_at(self) -> int:
        return self._callback_dict.get("created_at")

    @property
    def group_id(self) -> str:
        return self._callback_dict.get("group_id")

    @property
    def id(self) -> str:
        return self._callback_dict.get("id")

    @property
    def name(self) -> str:
        return self._callback_dict.get("name")

    @property
    def sender_id(self) -> str:
        return self._callback_dict.get("sender_id")

    @property
    def sender_type(self) -> str:
        return self._callback_dict.get("sender_type")

    @property
    def source_guid(self) -> str:
        return self._callback_dict.get("source_guid")

    @property
    def system(self) -> bool:
        return self._callback_dict.get("system")

    @property
    def text(self) -> str:
        return self._callback_dict.get("text")

    @property
    def user_id(self) -> str:
        return self._callback_dict.get("user_id")



import atexit
from typing import List, Dict

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from starlette.requests import Request
from starlette.responses import PlainTextResponse, JSONResponse
from starlette.types import Scope, Receive, Send, ASGIApp

from .bot import Bot

GET = 'GET'
POST = 'POST'
HEAD = 'HEAD'

_not_allowed = PlainTextResponse('405 Method Not Allowed', status_code=405)
_not_found = PlainTextResponse('404 Not Found', status_code=404)
_ping_handler = PlainTextResponse('Hello', status_code=200)


class RouteExistsError(Exception):
    pass


class Application(object):
    __slots__ = ('_scheduler', '_route_tree')
    _reserved_routes = ('/', '/_health')

    def __init__(self):
        """
        The Router is the primary object used to run the GroupMe Bot. Multiple Bots can be handled in one single
        router object. Each bot is assigned an endpoint path and requests to that endpoint will be handled by the
        associated bot.
        """
        self._scheduler: AsyncIOScheduler = AsyncIOScheduler()

        async def _summary(scope: Scope, receive: Receive, send: Send):
            response = JSONResponse({
                'endpoints': self.endpoints,
                'jobs': self.jobs,
                'scheduler_running': self._scheduler.running
            })
            await response(scope, receive, send)

        async def _health(scope: Scope, receive: Receive, send: Send):
            response = PlainTextResponse('OK')
            await response(scope, receive, send)

        self._route_tree: Dict[str, Dict[str, ASGIApp]] = {
            '/': {GET: _summary, HEAD: _ping_handler},
            '/_health': {GET: _health}
        }

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        req = Request(scope, receive, send)
        path = self._route_tree.get(req.url.path)
        if not path:
            await _not_found(scope, receive, send)
            return
        handler = path.get(req.method)
        if not handler:
            await _not_allowed(scope, receive, send)
            return
        await handler(scope, receive, send)

    def _start_scheduler(self):
        atexit.register(lambda: self._scheduler.shutdown(wait=False))
        self._scheduler.start()

    @property
    def routes(self) -> Dict[str, Dict[str, ASGIApp]]:
        """
        The current store of routes and the associated bot.
        :return dict[str, Bot]: key=the endpoint route, value=the bot object associated
        """
        return self._route_tree

    @property
    def scheduler(self) -> AsyncIOScheduler:
        """
        The AsyncIOScheduler object being used for cron jobs
        :return AsyncIOScheduler:
        """
        return self._scheduler

    @property
    def endpoints(self) -> Dict[str, Dict[str, str]]:
        """
        A summary of all endpoints within the Router
        """
        endpoint_summary = {}
        for path, methods in self._route_tree.items():
            for method, handler in methods.items():
                str_repr = handler.__name__ if hasattr(handler, '__name__') else str(handler)
                if endpoint_summary.get('path'):
                    endpoint_summary[path][method] = str_repr
                else:
                    endpoint_summary[path] = str_repr
        return endpoint_summary

    @property
    def jobs(self) -> List[str]:
        """
        A list of all currently running scheduler jobs
        :return List[str]: A list of the string repr of each running job
        """
        job_summary = []
        jobs = self._scheduler.get_jobs()
        if jobs:
            for job in jobs:
                job_summary.append(str(job))
        return job_summary

    def add_bot(self, bot: Bot, callback_path: str) -> None:
        """
        Add a new bot to be run by the Router
        :param bot: The bot to be run
        :param callback_path: The callback path for which the bot can be accesses
        :return:
        """
        # perform validity checks
        if callback_path in self._reserved_routes:
            raise RouteExistsError(f'Cannot use one of the reserved routes. Reserved '
                                   f'routes are: {str(self._reserved_routes)}')

        if callback_path in self._route_tree:
            raise RouteExistsError(f"Callback path `{callback_path}` is already in use. "
                                   f"You must use a new route for each bot.")

        # store the bot for call routing
        self._route_tree[callback_path] = {POST: bot, GET: _ping_handler, HEAD: _ping_handler}

        # add any scheduler jobs
        for job in bot.cron_jobs:
            self._scheduler.add_job(
                job['func'],
                job['trigger'],
                args=job['args'],
                **job['kwargs']
            )
            if not self._scheduler.running:
                self._start_scheduler()



from typing import List, Union, Dict


class InvalidAttachment(Exception):
    pass


class Attachment(object):
    __slots__ = '_kwargs'
    _list_keys = ('charmap', 'loci', 'user_ids')

    def __init__(self, **kwargs):
        self._kwargs: dict = kwargs

    @property
    def type(self):
        return self._kwargs.get('type')

    def to_dict(self) -> Dict[str, Union[str, List]]:
        """
        Converts the class instance to a dictionary with properly type casted values
        :return: A dict containing the properly type casted values
        """
        out = {}
        for key, value in self._kwargs.items():
            if key in self._list_keys:
                out[key] = value if value else []
            else:
                out[key] = str(value)
        return out


class ImageAttachment(Attachment):
    def __init__(self, image_url: str):
        Attachment.__init__(self, type='image', url=image_url)

    @property
    def url(self):
        return self._kwargs.get('url')


class LocationAttachment(Attachment):
    def __init__(self, name: str, lat: Union[str, float, int], lng: Union[str, float, int]):
        Attachment.__init__(self, type='location', name=name, lat=lat, lng=lng)

    @property
    def name(self):
        return self._kwargs.get('name')

    @property
    def lat(self):
        return self._kwargs.get('lat')

    @property
    def lng(self):
        return self._kwargs.get('lng')


class SplitAttachment(Attachment):
    def __init__(self, token: str):
        Attachment.__init__(self, type='split', token=token)

    @property
    def token(self):
        return self._kwargs.get('token')


class EmojiAttachment(Attachment):
    def __init__(self, placeholder: str, charmap: List[List[Union[int, int]]]):
        Attachment.__init__(self, type='emoji', placeholder=placeholder, charmap=charmap)

    @property
    def placeholder(self):
        return self._kwargs.get('placeholder')


class MentionsAttachment(Attachment):
    def __init__(self, loci: List[List[int]], user_ids: List[int]):
        Attachment.__init__(self, type='mentions', loci=loci, user_ids=user_ids)

    @property
    def loci(self):
        return self._kwargs.get('loci')

    @property
    def user_ids(self):
        return self._kwargs.get('user_ids')


def parse_attachment(attachment_dict: dict) -> Attachment:
    attachment_type = attachment_dict.get('type')
    if attachment_type is None:
        raise InvalidAttachment('`type` key not present')

    if attachment_type == 'image':
        return ImageAttachment(attachment_dict.get('url'))
    elif attachment_type == 'location':
        return LocationAttachment(attachment_dict.get('name'), attachment_dict.get('lat'), attachment_dict.get('lng'))
    elif attachment_type == 'split':
        return SplitAttachment(attachment_dict.get('token'))
    elif attachment_type == 'emoji':
        return EmojiAttachment(attachment_dict.get('placeholder'), attachment_dict.get('charmap'))
    elif attachment_type == 'mentions':
        return MentionsAttachment(attachment_dict.get('loci'), attachment_dict.get('user_ids'))

    raise InvalidAttachment(f'unsupported attachment type `{attachment_type}`')



import httpx

_status_codes = {
    200: "Success!",
    201: "Resource was created successfully.",
    204: "Resource was deleted successfully.",
    304: "There was no new data to return.",
    400: "Returned when an invalid format or invalid data is specified in the request.",
    401: "Authentication credentials were missing or incorrect.",
    403: "The request is understood, but it has been refused.",
    404: "The URI requested is invalid or the resource requested, such as a user, does not exists.",
    420: "You are being rate limited. Chill the heck out.",
    500: "Something unexpected occurred. GroupMe will be notified.",
    502: "GroupMe is down or being upgraded.",
    503: "The GroupMe servers are up, but overloaded with requests. Try again later."
}


def status_code_message(code: int):
    return _status_codes.get(code, "Unknown status code")


class GroupMeException(Exception):
    pass


class GroupMe(object):
    __slots__ = 'groupme_api_token'

    def __init__(self, groupme_api_token: str):
        self.groupme_api_token: str = groupme_api_token

    def image_url_to_groupme_image_url(self, image_url: str) -> str:
        """
        Convert a normal image URL to a GroupMe image
        :param str image_url: The URL for any image
        :return str: The URL for the converted GroupMe image
        """
        with httpx.Client() as client:
            res = client.get(image_url)
            res.raise_for_status()
            headers = {
                'X-Access-Token': self.groupme_api_token,
                'Content-Type': res.headers['Content-type'],
            }
            res = client.post(
                'https://image.groupme.com/pictures',
                headers=headers,
                content=res.content
            )
            res.raise_for_status()
            res = res.json()
            return res['payload']['picture_url']

    def get_group(self, group_id: str) -> dict:
        """
        Get a summary of the group from the GroupMe API
        :return dict:
        """
        return self.__get(f'/groups/{group_id}')

    def __get(self, path: str) -> dict:
        res = httpx.get(
            f'https://api.groupme.com/v3{path}',
            params={'token': self.groupme_api_token}
        )
        res.raise_for_status()
        res = res.json()
        status_code = res['meta']['code']
        if status_code <= 299:
            return res['response']
        elif status_code >= 400:
            msg = status_code_message(status_code)
            errors = res.get('meta', {}).get('errors', '')
            raise GroupMeException(f"{status_code} {msg} {str(errors)}")



from __future__ import annotations

import re
from collections import OrderedDict
from json.decoder import JSONDecodeError
from typing import Any, List, Callable, Optional

import httpx
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.types import Scope, Receive, Send

from .attachment import Attachment, MentionsAttachment
from .callback import Callback
from .groupme import GroupMe

_success_response = PlainTextResponse('Success')


class HandlerPatternExistsError(Exception):
    pass


class Context(object):
    __slots__ = ('_bot', '_callback')

    def __init__(self, bot: Bot, callback: Callback):
        """
        Context provided to every handler/scheduled function run by the bot. This provides a clean object containing
        bot the bot in use and the Callback that was sent.
        :param Bot bot:
        :param Callback callback:
        """
        self._bot: Bot = bot
        self._callback: Callback = callback

    @property
    def bot(self) -> Bot:
        return self._bot

    @property
    def callback(self) -> Callback:
        return self._callback


class Bot(GroupMe):
    __slots__ = ('bot_name', 'bot_id', 'groupme_api_token', 'group_id', '_handler_functions', '_jobs')

    def __init__(self, bot_name: str, bot_id: str, groupme_api_token: str, group_id: str):
        """
        The Bot class represents a single bot that can contains multiple callback handlers and scheduled jobs.
        Bots are run using the Router class.
        :param bot_name: A unique name for this bot to be known as within this code. Does not impact the name
            displayed in the GroupMe app.
        :param bot_id: The Bot ID provided by GroupMe.
        :param groupme_api_token: The GroupMe API token for access to group details.
        :param group_id: The id of the GroupMe group in which the bot exists.
        """
        super().__init__(groupme_api_token)
        self.bot_name = bot_name
        self.bot_id = bot_id
        self.groupme_api_token = groupme_api_token
        self.group_id = group_id

        self._handler_functions: OrderedDict = OrderedDict()
        self._jobs = []

    @property
    def cron_jobs(self) -> List[dict]:
        """
        All list of cron jobs associated with this bot.
        :return List[dict]:
        """
        return self._jobs

    def __str__(self):
        return f"{self.bot_name}: {len(self._handler_functions)} callback handlers, " \
               f"{len(self._jobs)} cron jobs at {hex(id(self))}"

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        req = Request(scope, receive, send)
        try:
            callback_dict = await req.json()
        except JSONDecodeError as e:
            response = PlainTextResponse(
                "400 Bad Request. Unable to parse JSON. Error: " + str(e), status_code=400)
            await response(scope, receive, send)
            return
        callback = Callback(callback_dict)
        if callback.sender_type != 'user':  # only reply to users
            await _success_response(scope, receive, send)
            return
        text = callback.text.lower().strip()
        for pattern, func in self._handler_functions.items():
            if re.search(pattern, text):
                try:
                    func(Context(self, callback))
                    await _success_response(scope, receive, send)
                    return
                except Exception as e:
                    response = PlainTextResponse(str(e), status_code=500)
                    await response(scope, receive, send)
                    return
        await _success_response(scope, receive, send)
        return

    def add_callback_handler(self, regex_pattern: str, func: Callable[[Context], Any]) -> None:
        """
        Registers a regex pattern as to a bot handler function. If the regex pattern
        is found in a message from a GroupMe user, the function will be called.
        :param regex_pattern: The pattern to search for in the message text
        :param Callable[[Context], Any] func: The function to be called when the pattern is matched
        """
        if regex_pattern in self._handler_functions:
            raise HandlerPatternExistsError(f"The pattern `{regex_pattern}` is already registered to a handler")
        self._handler_functions[regex_pattern] = func

    def add_cron_job(self, func: Callable[[Context], Any], **kwargs) -> None:
        """
        Registers a function to be run on set cron schedule.
        Uses APScheduler cron trigger. Details here:
            https://apscheduler.readthedocs.io/en/stable/modules/triggers/cron.html
        Available arguments:
            - year (int|str) – 4-digit year
            - month (int|str) – month (1-12)
            - day (int|str) – day of the (1-31)
            - week (int|str) – ISO week (1-53)
            - day_of_week (int|str) – number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun)
            - hour (int|str) – hour (0-23)
            - minute (int|str) – minute (0-59)
            - second (int|str) – second (0-59)
            - start_date (datetime|str) – earliest possible date/time to trigger on (inclusive)
            - end_date (datetime|str) – latest possible date/time to trigger on (inclusive)
            - timezone (datetime.tzinfo|str) – time zone to use for the date/time calculations
                (defaults to scheduler timezone)
            - jitter (int|None) – advance or delay the job execution by jitter seconds at most.
        :param Callable[[Context], Any] func: The function to be called when the cron trigger is triggered
        """
        self._jobs.append({
            'func': func,
            'trigger': 'cron',
            'args': (Context(self, Callback({})),),
            'kwargs': kwargs
        })

    def post_message(self, msg: str, attachments: Optional[List[Attachment]] = None) -> httpx.Response:
        """
        Posts a bot message to the group with optional attachments.
        :param str msg: The message to be sent
        :param Optional[List[Attachment]] attachments: Attachments to send in the message
        :return requests.Response: The POST request response object
        """
        if attachments:
            attachments = [attachment.to_dict() for attachment in attachments]
        else:
            attachments = []
        data = {
            "bot_id": self.bot_id,
            "text": msg,
            "attachments": attachments
        }
        response = httpx.post(
            'https://api.groupme.com/v3/bots/post',
            json=data,
            headers={'Content-Type': 'application/json'}
        )
        response.raise_for_status()
        return response

    def mention_all(self) -> None:
        """
        Mentions everybody in the group so they receive a notification
        """
        text = ''
        user_ids = []
        loci = []
        group = self.get_group(self.group_id)
        for member in group['members']:
            user_ids.append(member['user_id'])
            loci.append([len(text), len(member['nickname']) + 1])
            text += '@{} '.format(member['nickname'])
        self.post_message(text, [MentionsAttachment(loci=loci, user_ids=user_ids)])
